# Executables in the `bin` directory 

The `bin` directory of a GraalVM contains two categories of "launchers".
This file explains the requirements and architecture of the second category of those.

* Executables (Native Images or Bash scripts) generated by GraalVM launch a language's compiler/interpreter. This enables us to run `ruby
  <args>` rather than something like `java <java options> <java classpath> <main class> <args>`.
* Additional scripts which launch related command-line tools like `gem` and `irb`,
  which are from the `bin` directory of the `truffleruby` repository. This is the topic of this document.

## Requirements

1.  When TruffleRuby tar (GraalVM) is downloaded and unpacked, all files in `bin` and 
    `jre/languages/ruby/bin` has to always run the TruffleRuby provided by the tar. 
    -   Therefore `PATH` is not involved (`/usr/bin/env` cannot be used in shebang).
    -   Therefore there are no absolute paths in `bin` executables.
    -   Therefore all scripts in `bin` are always resolved to `truffleruby` in the same dir.
2.  Newly installed gem executables has to always run the `truffleruby` from the same `bin`. 
    -   RubyGems generate executables with absolute path by default.
3.  Behave as proper Ruby implementation when the `bin` is put on `PATH`.
    -   Works.
4.  RubyMine requires `gem` in `bin` to work when executed with `truffleruby -x` option 
    (otherwise gems are not listed).
    -   Therefore `-x` option is implemented.
5.  `ruby -S irb` when executed in a `bin` directory has to work. (Applies to other 
    executables in `bin` as well.)
    -   When a file starting with a shebang not containing `ruby` is loaded the lines at 
        the begging are skipped until a shebang containing `'ruby'` is found (it has `-x` 
        option behaviour). This is the behaviour of MRI and Truffleruby implements it as well.
6.  Any executable has to work when pwd is `bin` directory.
    -   Works.
7.  Works on Linux, Solaris, macOS.
    -   Works.
8.  Works with Ruby managers. (Stacked shebangs cause problems, since the `execve` can 
    resolve only one level)
    -   Therefore `truffleruby` is a binary executable (at least on macOS).
    -   Issues could otherwise be observed with rbenv on macOS with fish shell.

## Solution

Truffleruby uses hybrid executables, an example of `gem` file follows:

```ruby
#!/usr/bin/env bash
# ignored by Ruby interpreter

# get the absolute path of the executable and resolve symlinks
SELF_PATH=$(cd "$(dirname "$0")" && pwd -P)/$(basename "$0")
while [ -h "$SELF_PATH" ]; do
  # 1) cd to directory of the symlink
  # 2) cd to the directory of where the symlink points
  # 3) get the pwd
  # 4) append the basename
  DIR=$(dirname "$SELF_PATH")
  SYM=$(readlink "$SELF_PATH")
  SELF_PATH=$(cd "$DIR" && cd "$(dirname "$SYM")" && pwd)/$(basename "$SYM")
done
exec "$(dirname $SELF_PATH)/ruby" "$SELF_PATH" "$@"

#!ruby
# ^ marks start of Ruby interpretation, content of current lib/bin follows   

#--
# Copyright 2006 by Chad Fowler, Rich Kilmer, Jim Weirich and others.
# All rights reserved.
# See LICENSE.txt for permissions.
#++

require 'rubygems'
require 'rubygems/gem_runner'
require 'rubygems/exceptions'

required_version = Gem::Requirement.new ">= 1.8.7"

unless required_version.satisfied_by? Gem.ruby_version then
  abort "Expected Ruby Version #{required_version}, is #{Gem.ruby_version}"
end

args = ARGV.clone

begin
  Gem::GemRunner.new.run args
rescue Gem::SystemExitException => e
  exit e.exit_code
end
```

If executed directly, bash interprets the file. It looks up correct ruby
executable  (`truffleruby` in the same directory) and executes the same file
with it. When the Ruby interpreter executes the file it ignores the first 2 lines
running only the Ruby portion of the file.
