/*
 * Copyright (c) 2015, 2019 Oracle and/or its affiliates. All rights reserved. This
 * code is released under a tri EPL/GPL/LGPL license. You can use it,
 * redistribute it and/or modify it under the terms of the:
 *
 * Eclipse Public License version 2.0, or
 * GNU General Public License version 2, or
 * GNU Lesser General Public License version 2.1.
 */
package org.truffleruby.core.mutex;

import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

import com.oracle.truffle.api.nodes.Node;
import org.truffleruby.Layouts;
import org.truffleruby.RubyContext;
import org.truffleruby.core.exception.ExceptionOperations;
import org.truffleruby.core.thread.ThreadManager;
import org.truffleruby.language.RubyNode;
import org.truffleruby.language.control.RaiseException;

import com.oracle.truffle.api.CompilerDirectives.TruffleBoundary;
import com.oracle.truffle.api.object.DynamicObject;
import com.oracle.truffle.api.profiles.BranchProfile;

public abstract class MutexOperations {

    @TruffleBoundary
    protected static void lock(RubyContext context, ReentrantLock lock, DynamicObject thread, RubyNode currentNode) {
        lockInternal(context, lock, currentNode);
        Layouts.THREAD.getOwnedLocks(thread).add(lock);
    }

    @TruffleBoundary
    public static void lockInternal(RubyContext context, ReentrantLock lock, RubyNode currentNode) {
        if (lock.tryLock()) {
            return;
        }

        context.getThreadManager().runUntilResult(currentNode, () -> {
            lock.lockInterruptibly();
            return ThreadManager.BlockingAction.SUCCESS;
        });
    }

    @TruffleBoundary
    protected static void lockEvenWithExceptions(
            RubyContext context, ReentrantLock lock, DynamicObject thread, Node currentNode) {
        // We need to re-lock this lock after a Mutex#sleep, no matter what, even if another thread throw us an exception.
        // Yet, we also need to allow safepoints to happen otherwise the thread that could unlock could be blocked.
        try {
            internalLockEvenWithException(context, lock, currentNode);
        } finally {
            Layouts.THREAD.getOwnedLocks(thread).add(lock);
        }
    }

    protected static void internalLockEvenWithException(RubyContext context, ReentrantLock lock, Node currentNode) {
        if (lock.isHeldByCurrentThread()) {
            throw new RaiseException(context, context.getCoreExceptions().threadErrorRecursiveLocking(currentNode));
        }

        if (lock.tryLock()) {
            return;
        }
        Throwable throwable = null;
        try {
            while (true) {
                try {
                    context.getThreadManager().runUntilResult(currentNode, () -> {
                        lock.lockInterruptibly();
                        return ThreadManager.BlockingAction.SUCCESS;
                    });
                    break;
                } catch (Throwable t) {
                    throwable = t;
                }
            }
        } finally {
            if (!lock.isHeldByCurrentThread()) {
                throw new AssertionError("the lock could not be reacquired after Mutex#sleep");
            }
        }

        if (throwable != null) {
            ExceptionOperations.rethrow(throwable);
        }
    }

    @TruffleBoundary
    protected static void unlock(ReentrantLock lock, DynamicObject thread) {
        unlockInternal(lock);
        Layouts.THREAD.getOwnedLocks(thread).remove(lock);
    }

    @TruffleBoundary
    public static void unlockInternal(ReentrantLock lock) {
        assert lock.isHeldByCurrentThread();
        lock.unlock();
    }

    @TruffleBoundary
    public static Condition newCondition(ReentrantLock lock) {
        return lock.newCondition();
    }

    @TruffleBoundary
    public static void checkOwnedMutex(RubyContext context, ReentrantLock lock, RubyNode currentNode,
            BranchProfile errorProfile) {
        if (!lock.isHeldByCurrentThread()) {
            errorProfile.enter();
            if (!lock.isLocked()) {
                throw new RaiseException(context, context.getCoreExceptions().threadErrorUnlockNotLocked(currentNode));
            } else {
                throw new RaiseException(context, context.getCoreExceptions().threadErrorAlreadyLocked(currentNode));
            }
        }
    }

}
